/* > rodate.c
 *
 * Get RISC OS date for use with earlier versions of !PC, !PCx86 and !PCPro
 * that do not correctly pass the date after 7 Dec 1999/1 Jan 2000.
 *
 * (c) B.A. Brand, 2000
 * version 1.01, 16 Jan 2000
 *
 * Requires ARMEDIT.LIB by Alexander Thoukydides.
 *
 * History:
 * 1.01, 16 Jan 2000: Honour local country settings when displaying date
 *                    and time.  Give a bit more help and information.
 * 1.00,  9 Jan 2000: First version.
 *
 */

/* Note: Compile with tiny memory model (the ARMEDIT library is compiled
   with that model). */
#ifndef __TINY__
#error Compile with tiny memory model!
#endif

#include <dos.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "talk.h"

const char * basename(const char *);
void usage(void);
void t_error(const talk_error *);
void showdatetime(struct dosdate_t *,struct dostime_t *);

/* Inline this function in C++. */
#ifdef __cplusplus
inline
#endif
const talk_error * bb_talk_pokeb(unsigned long addr,char b) {
  return talk_write(&b,1,addr);
}

/* Program name, global for error messages. */
char * myname = NULL;

/* Address of RISC OS memory buffer, global for error clean-up. */
long ro_buf = 0;


int main(int argc, char ** argv) {
  /* Use local territory or UTC? */
  int utc = 0;
  /* talk_...() error pointer. */
  const talk_error * te;
  /* SWI register structures. */
  talk_swi_regs rin,rout;
  /* DOS copy of RISC OS memory buffer */
  unsigned char ro_buf_copy[32];
  /* DOS date and time structures. */
  struct dosdate_t dosdate;
  struct dostime_t dostime;

  /* Get name of program invocation.
     Note: strdup malloc()s its own space. */
  myname = strdup(basename(argv[0]));
  /* Strip .com or .exe. */
  if (myname) {
    size_t s;
    if ((s = strlen(myname)) > 4) {
      myname[s-4] = '\0';
    }
  }
  else {
    myname = "RODATE"; /* default name */
  }

  /* Quick check of command line arguments. */
  switch (argc) {
    case 1: utc = 0; break;
    case 2: if (0 == stricmp(argv[1],"/u") ||
                0 == stricmp(argv[1],"-u")) {
              utc = 1;
              break;
            }
            /* else fall through */
    default: usage(); break;
  }

  /* General opening message. */
  printf("%s\n(c) B.A. Brand, 2000, using ARMEDIT (c) A. Thoukydides, 1997\n"
         "Setting DOS date and time to RISC OS %s date and time.\n\n",
         myname,(utc ? "UTC" : "local"));

  /* Set buffer for SWI calls, one buffer should suffice.
     5 bytes for OS_Word, round to 8.
     32 bytes for Territory_ConvertTimeToOrdinals or
     Territory_ConvertTimeToUTCOrdinals. */
  if (NULL != (te = talk_malloc(40,&ro_buf))) t_error(te);

  /* Get RISC OS time (5 byte unsigned integer, centiseconds since
     1 Jan 1900, 00:00:00.00).
     Call OS_Word 14,3. */
  memset(&rin ,0,sizeof(rin )); /* clear memory */
  rin.r[0] = 14L;
  rin.r[1] = ro_buf;
  if (NULL != (te = bb_talk_pokeb(ro_buf,3)))  t_error(te);
  if (NULL != (te = talk_swi(0x7,&rin,&rout))) t_error(te);

  /* Convert the date and time to DOS format.
     talk_date_to_dos is not suited, this converts to the format used for
     file time and date.
     So use RISC OS to do the conversion.  Definitely do NOT try to do the
     calculations oneself! */
  memset(&rin, 0,sizeof(rin ));
  rin.r[1] = ro_buf   ;
  rin.r[2] = ro_buf+8L;
  if (utc) {
    /* Get UTC time through Territory_ConvertTimeToUTCOrdinals. */
    if (NULL != (te = talk_swi(0x43049,&rin,&rout))) t_error(te);
  }
  else {
    /* Get local time through Territory_ConvertTimeToOrdinals. */
    rin.r[0] = -1L; // use local territory.
    if (NULL != (te = talk_swi(0x4304f,&rin,&rout))) t_error(te);
  }

  /* And set DOS clock.
     Use Microsoft compatible (??) functions _dos_getdate(), _dos_gettime(),
     _dos_setdate() and _dos_settime() rather than Borland's getdate() etc.
     If neither of these functions is present in the library of your compiler
     use INT 21 functions 2A to 2D, see Ralph Brown's Interrupt list. */

  _dos_getdate(&dosdate);
  _dos_gettime(&dostime);

  printf("Current DOS date and time: ");
  showdatetime(&dosdate,&dostime);
  printf("\n");

  /* Copy ro_buf to DOS side, better than 7 separate reads. */
  if ((te = talk_read(ro_buf_copy,sizeof(ro_buf_copy),ro_buf+8)) != NULL)
    t_error(te);
  dostime.hsecond = *(unsigned long*)  ro_buf_copy    ;
  dostime.second  = *(unsigned long*) (ro_buf_copy+4 );
  dostime.minute  = *(unsigned long*) (ro_buf_copy+8 );
  dostime.hour    = *(unsigned long*) (ro_buf_copy+12);
  dosdate.day     = *(unsigned long*) (ro_buf_copy+16);
  dosdate.month   = *(unsigned long*) (ro_buf_copy+20);
  dosdate.year    = *(unsigned long*) (ro_buf_copy+24);
  /* day_of_week field not used on setting date, but it is in
     *(unsigned long*) (ro_buf_copy+28) - 1. */
  _dos_setdate(&dosdate);
  _dos_settime(&dostime);

  /* Well, just a check... so reread structures after setting */
  _dos_getdate(&dosdate);
  _dos_gettime(&dostime);
  printf("New date & time from RISC OS: ");
  showdatetime(&dosdate,&dostime);
  if (utc) printf(" UTC");
  printf("\n");

  /* Release RISC OS memory. */
  talk_free(ro_buf);

  return EXIT_SUCCESS;
}


/*
 * Strip directory part of file name.
 * (More elaborate version may strip .com or .exe as well, but this is a
 * quick version since Borland 3.1 does not have basename() in its standard
 * libraries.)
 */
const char * basename(const char * s) {
  const char * cp;
  if (!s) return NULL;

  cp = s;
  while (*cp++); cp--; /* strlen() */
  /* DOS may use both / and \ internally!
     Only COMMAND.COM insists on \ as directory separators. */
  while (cp >= s && *cp != '/' && *cp != '\\' && *cp != ':') cp--;
  return cp+1;
}


/*
 * Give usage message.
 * (May use a screenwidth-aware pretty-printer if one is present).
 */
void usage(void) {
  printf("Usage: %s [/u]\n\n"
"This program gets the date and time from the RISC OS side of the computer and"
"\n"
"sets it in this DOS session using the currently configured RISC OS territory,"
"\n"
"timezone and daylight saving time/summertime setting."
"\n"
"The optional /u (or -u) switch gets the UTC time from RISC OS, irrespective"
"\n"
"of territory setting.  Useful if you want to use Windows 9x's own timezone\n"
"and daylight saving time/summertime management."
"\n\n"
"(c) B.A. Brand, 2000.  Requires ARMEDIT (c) A. Thoukydides, 1997\n"
    ,myname);

  /* Extremely paranoid programmers would add a talk_free() here too.
     And usage is not strictly an error. */
  exit(EXIT_FAILURE);
}


/*
 * React to error message from any of ARMEDIT's talk functions
 * and free talk_malloc()'d memory.
 */
void t_error(const talk_error * te) {
  fprintf(stderr,"%s: RISC OS or ARMEDIT error `%s\'.\n",
                 myname,te->errmess);
  if (ro_buf != 0) {
    talk_free(ro_buf);
    ro_buf = 0;
  }
  exit(EXIT_FAILURE);
}


/*
 * Display date and time in dosdate_t and dostime_t format honouring the
 * current country settings (if possible).
 * Country settings from INT 21 function 38 if your compiler does not have
 * a country() function.
 */
void showdatetime(struct dosdate_t * dosdate, struct dostime_t * dostime) {
  static struct COUNTRY c_info; /* rather than global, need to read it once */
  static int c_info_read = 0;
  unsigned hour;

  if (! c_info_read) {
    if (NULL == country(0,&c_info)) {
      /* No country information available, use `default' European style date
         and 24 hour clock.  (After all, this is a `continental' program.) */
       c_info.co_date = c_info.co_time = 1;
       strcpy(c_info.co_dtsep,"-");
       strcpy(c_info.co_tmsep,":");
       strcpy(c_info.co_desep,",");
    }
    c_info_read = 1;
  }

  /* Date. */
  switch (c_info.co_date) {
    case 0: /* USA: mm-dd-yyyy */
      printf("%02d%s%02d%s%04d",dosdate->month,c_info.co_dtsep,
                                dosdate->day,  c_info.co_dtsep,
                                dosdate->year);
      break;
    case 1: /* European: dd-mm-yyyy */
      printf("%02d%s%02d%s%04d",dosdate->day,  c_info.co_dtsep,
                                dosdate->month,c_info.co_dtsep,
                                dosdate->year);
      break;
    case 2: /* ISO/Japanese: yyyy-mm-dd */
    default:
      printf("%04d%s%02d%s%02d",dosdate->year, c_info.co_dtsep,
                                dosdate->month,c_info.co_dtsep,
                                dosdate->day);
      break;
  }

  /* Time. */
  if (c_info.co_time & 1) {
    /* 24 hour clock */
    hour = dostime->hour;
  }
  else {
    /* 12 hour AM/PM clock */
    hour = dostime->hour % 12;
    if (0 == hour) hour += 12;
  }
  printf(" %02d%s%02d%s%02d%s%02d",hour,           c_info.co_tmsep,
                                   dostime->minute,c_info.co_tmsep,
                                   dostime->second,c_info.co_desep,
                                   dostime->hsecond);
  if (0 == (c_info.co_time & 1))
    printf(" %cM",(char)(dostime->hour / 12 ? 'P' : 'A'));
}
